# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
from os.path import join, expanduser, isdir
from ovos_config.locations import get_xdg_data_save_path
from ovos_config.meta import get_xdg_base


class FileSystemAccess:
    """A class for providing access to the mycroft FS sandbox.

    Intended to be attached to skills at initialization time to provide a
    skill-specific namespace.
    """

    def __init__(self, path):
        #: Member value containing the root path of the namespace
        self.path = self.__init_path(path)

    @staticmethod
    def __init_path(path):
        if not isinstance(path, str) or len(path) == 0:
            raise ValueError("path must be initialized as a non empty string")

        old_path = expanduser(f'~/.{get_xdg_base()}/{path}')
        xdg_path = expanduser(f'{get_xdg_data_save_path()}/filesystem/{path}')
        # Migrate from the old location if it still exists
        if isdir(old_path) and not isdir(xdg_path):
            shutil.move(old_path, xdg_path)

        if not isdir(xdg_path):
            os.makedirs(xdg_path)
        return xdg_path

    def open(self, filename, mode):
        """Open a file in the provided namespace.

        Get a handle to a file (with the provided mode) within the
        skill-specific namespace.

        Parameters:
            filename (str): a path relative to the namespace.
                      subdirs not currently supported.

            mode (str): a file handle mode

        Returns:
            an open file handle.
        """
        file_path = join(self.path, filename)
        return open(file_path, mode)

    def exists(self, filename):
        """Check if file exists in the namespace.

        Args:
            filename (str): a path relative to the namespace.
                      subdirs not currently supported.
        Returns:
            bool: True if file exists, else False.
        """
        return os.path.exists(join(self.path, filename))



from os.path import isdir, join

from ovos_config.locations import get_xdg_config_save_path
from ovos_utils.messagebus import get_mycroft_bus
from ovos_utils.log import LOG

from ovos_workshop.resource_files import locate_lang_directories
from ovos_workshop.skills.ovos import OVOSSkill


class OVOSAbstractApplication(OVOSSkill):
    def __init__(self, skill_id, bus=None, resources_dir=None,
                 lang=None, settings=None, gui=None, enable_settings_manager=False):
        super().__init__(bus=bus, gui=gui, resources_dir=resources_dir,
                         enable_settings_manager=enable_settings_manager)
        self.skill_id = skill_id
        self._dedicated_bus = False
        if bus:
            self._dedicated_bus = False
        else:
            self._dedicated_bus = True
            bus = get_mycroft_bus()
        self._startup(bus, skill_id)
        if settings:
            LOG.warning("settings arg is deprecated and will be removed "
                        "in a future release")
            self.settings.merge(settings)

    @property
    def _settings_path(self):
        return join(get_xdg_config_save_path(), 'apps', self.skill_id,
                    'settings.json')

    def default_shutdown(self):
        self.clear_intents()
        super().default_shutdown()
        if self._dedicated_bus:
            self.bus.close()

    def get_language_dir(self, base_path=None, lang=None):
        """ checks for all language variations and returns best path
        eg, if lang is set to pt-pt but only pt-br resources exist,
        those will be loaded instead of failing, or en-gb vs en-us and so on
        """

        base_path = base_path or self.res_dir
        lang = lang or self.lang
        lang_path = join(base_path, lang)

        # base_path/en-us
        if isdir(lang_path):
            return lang_path

        # check for subdialects of same language as a fallback
        # eg, language is set to en-au but only en-us resources are available
        similar_dialect_directories = locate_lang_directories(lang, base_path)
        for directory in similar_dialect_directories:
            if directory.exists():
                return directory

    def clear_intents(self):
        # remove bus handlers, otherwise if re-registered we get multiple
        # handler executions
        for intent_name, _ in self.intent_service:
            event_name = f'{self.skill_id}:{intent_name}'
            self.remove_event(event_name)

        self.intent_service.detach_all()  # delete old intents before re-registering



from ovos_workshop.skills.ovos import OVOSSkill


class ActiveSkill(OVOSSkill):
    def bind(self, bus):
        super(ActiveSkill, self).bind(bus)
        if bus:
            """ insert skill in active skill list on load """
            self.make_active()

    def handle_skill_deactivated(self, message=None):
        """ skill is always in active skill list,
        ie, converse is always called """
        self.make_active()





from threading import Event
from time import time as get_time, sleep
from ovos_utils.log import LOG
from ovos_utils.messagebus import Message
from ovos_workshop.skills.ovos import OVOSFallbackSkill
from ovos_config.config import read_mycroft_config, update_mycroft_config


class BaseIntentEngine:
    # TODO move to OPM
    def __init__(self, name, config=None):
        self.name = name.lower()
        config = config or read_mycroft_config()
        self.config = config.get(self.name, {})
        self.intent_samples = {}
        self.entity_samples = {}
        self.regex_samples = {}

    def add_intent(self, name, samples):
        self.intent_samples[name] = samples

    def remove_intent(self, name):
        if name in self.intent_samples:
            del self.intent_samples[name]

    def add_entity(self, name, samples):
        self.entity_samples[name] = samples

    def add_regex(self, name, pattern):
        self.regex_samples[name] = pattern

    def remove_regex(self, name):
        if name in self.regex_samples:
            del self.regex_samples[name]

    def remove_entity(self, name):
        if name in self.entity_samples:
            del self.entity_samples[name]

    def train(self, single_thread=False):
        """ train all registered intents and entities"""
        pass

    def calc_intent(self, query):
        """ return best intent for this query  """
        data = {"conf": 0,
                "utterance": query,
                "name": None}
        return data


class IntentEngineSkill(OVOSFallbackSkill):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.engine = None
        self.config = {}
        self.priority = 1

    def initialize(self):
        engine = BaseIntentEngine("dummy")
        self.bind_engine(engine, self.priority)

    def bind_engine(self, engine, priority=4):
        priority_skills = read_mycroft_config().get("skills", {}).get(
            "priority_skills", [])
        priority_skills.append(self.root_dir.split("/")[-1])
        update_mycroft_config({
            "skills":
                {"priority_skills": priority_skills}
        })
        self.priority = priority
        self.engine = engine
        self.config = engine.config
        self.register_messages(engine.name)
        self.register_fallback(self.handle_fallback, self.priority)
        self.finished_training_event = Event()
        self.finished_initial_train = False
        self.train_delay = self.engine.config.get('train_delay', 4)
        self.train_time = get_time() + self.train_delay

    def register_messages(self, name):
        self.bus.on('mycroft.skills.initialized', self.train)
        self.bus.on(name + ':register_intent', self._register_intent)
        self.bus.on(name + ':register_entity', self._register_entity)
        self.bus.on(name + ':register_regex', self._register_regex)

    def train(self, message=None):
        single_thread = message.data.get('single_thread', False)
        self.finished_training_event.clear()

        LOG.info('Training...')
        self.engine.train(single_thread=single_thread)
        LOG.info('Training complete.')

        self.finished_training_event.set()
        self.finished_initial_train = True

    def wait_and_train(self):
        if not self.finished_initial_train:
            return
        sleep(self.train_delay)
        if self.train_time < 0.0:
            return

        if self.train_time <= get_time() + 0.01:
            self.train_time = -1.0
            self.train()

    def _register_object(self, message, object_name, register_func):
        name = message.data['name']
        samples = message.data['samples']

        LOG.debug(
            'Registering ' + self.engine.name + ' ' + object_name + ': ' + name)

        register_func(name, samples)
        self.train_time = get_time() + self.train_delay
        self.wait_and_train()

    def register_intent(self, name, samples):
        data = {"name": name, "samples": samples}
        self._register_intent(Message(name, data))

    def register_entity(self, name, samples):
        data = {"name": name, "samples": samples}
        self._register_entity(Message(name, data))

    def register_regex(self, name, pattern):
        data = {"name": name, "pattern": pattern}
        self._register_regex(Message(name, data))

    def _register_intent(self, message):
        self._register_object(message, 'intent', self.engine.add_intent)

    def _register_entity(self, message):
        self._register_object(message, 'entity', self.engine.add_entity)

    def _register_regex(self, message):
        self._register_object(message, 'regex', self.engine.add_regex)

    def handle_fallback(self, message):
        utt = message.data.get('utterance')
        LOG.debug(self.engine.name + " fallback attempt: " + utt)

        if not self.finished_training_event.is_set():
            LOG.debug('Waiting for training to finish...')
            self.finished_training_event.wait()

        data = self.engine.calc_intent(utt)

        if data["conf"] < 0.5:
            return False

        self.make_active()

        self.bus.emit(message.reply(data["name"], data=data))
        return True



from ovos_config import Configuration
from ovos_plugin_manager.language import OVOSLangDetectionFactory, OVOSLangTranslationFactory
from ovos_utils import get_handler_name
from ovos_utils.log import LOG
from ovos_workshop.resource_files import SkillResources
from ovos_workshop.skills.ovos import OVOSSkill, OVOSFallbackSkill


class UniversalSkill(OVOSSkill):
    ''' Skill that auto translates input/output from any language '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lang_detector = OVOSLangDetectionFactory.create()
        self.translator = OVOSLangTranslationFactory.create()

        self.internal_language = None  # the skill internally only works in this language
        self.translate_tags = True  # __tags__ private value will be translated (adapt entities)
        self.translate_keys = []  # any keys added here will have values translated in message.data
        if self.internal_language is None:
            lang = Configuration().get("lang", "en-us")
            LOG.warning(f"UniversalSkill are expected to specify their internal_language, casting to {lang}")
            self.internal_language = lang

    def _load_lang(self, root_directory=None, lang=None):
        """unlike base skill class all resources are in self.internal_language by default
        instead of self.lang (which comes from message)
        this ensures things like self.dialog_render reflect self.internal_lang
        """
        lang = lang or self.internal_language  # self.lang in base class
        root_directory = root_directory or self.res_dir
        if lang not in self._lang_resources:
            self._lang_resources[lang] = SkillResources(root_directory, lang, skill_id=self.skill_id)
        return self._lang_resources[lang]

    def detect_language(self, utterance):
        try:
            return self.lang_detector.detect(utterance)
        except:
            # self.lang to account for lang defined in message
            return self.lang.split("-")[0]

    def translate_utterance(self, text, target_lang, sauce_lang=None):
        sauce_lang = sauce_lang or self.detect_language(text)
        if sauce_lang.split("-")[0] != target_lang:
            translated = self.translator.translate(text, source=sauce_lang, target=target_lang)
            LOG.info("translated " + text + " to " + translated)
            return translated
        return text

    def _translate_message(self, message):
        # translate speech from input lang to internal lang
        sauce_lang = self.lang  # from message or config
        out_lang = self.internal_language  # skill wants input is in this language,

        ut = message.data.get("utterance")
        if ut:
            message.data["utterance"] = self.translate_utterance(ut, target_lang=out_lang, sauce_lang=sauce_lang)
        if "utterances" in message.data:
            message.data["utterances"] = [self.translate_utterance(ut, target_lang=out_lang, sauce_lang=sauce_lang)
                                          for ut in message.data["utterances"]]
        for key in self.translate_keys:
            if key in message.data:
                ut = message.data[key]
                message.data[key] = self.translate_utterance(ut, target_lang=out_lang, sauce_lang=sauce_lang)
        if self.translate_tags:
            for idx, token in enumerate(message.data["__tags__"]):
                message.data["__tags__"][idx] = self.translate_utterance(token.get("key", ""),
                                                                         target_lang=out_lang,
                                                                         sauce_lang=sauce_lang)
        return message

    def create_universal_handler(self, handler):

        def universal_intent_handler(message):
            message = self._translate_message(message)
            LOG.info(get_handler_name(handler))
            handler(message)

        return universal_intent_handler

    def register_intent(self, intent_parser, handler):
        handler = self.create_universal_handler(handler)
        super().register_intent(intent_parser, handler)

    def register_intent_file(self, intent_file, handler):
        handler = self.create_universal_handler(handler)
        super().register_intent_file(intent_file, handler)

    def speak(self, utterance, *args, **kwargs):
        # translate speech from input lang to output lang
        out_lang = self.lang  # from message or config
        sauce_lang = self.internal_language  # skill output is in this language
        utterance = self.translate_utterance(utterance, sauce_lang, out_lang)
        super().speak(utterance, *args, **kwargs)


class UniversalFallback(UniversalSkill, OVOSFallbackSkill):
    ''' Fallback Skill that auto translates input/output from any language '''

    def create_universal_fallback_handler(self, handler):
        def universal_fallback_handler(message):
            # auto_Translate input
            message = self._translate_message(message)
            LOG.info(get_handler_name(handler))
            success = handler(self, message)
            if success:
                self.make_active()
            return success

        return universal_fallback_handler

    def register_fallback(self, handler, priority):
        handler = self.create_universal_fallback_handler(handler)
        super().register_fallback(handler, priority)
